// Copyright (c) 2015, Herman Bergwerf. All rights reserved.
// Use of this source code is governed by a MIT-style license
// that can be found in the LICENSE file.

/// # Omnibus library
///
/// The flowchart below shows the architecture of the Omnibus. The Omnibus has
/// separate channels for each possible event. Anyone can receive events by
/// subscribing to the event channel in the Omnibus. Listeners can reply to an
/// event. All replies will be returned to the origin of the event.
///
/// ```
///           +---> Replies   Event A +---+
///           |                           |
///           |                           |
///           |                           |
///       +---+---+    +---------+    +---v--+
/// +-----> Reply +----> Omnibus <----+ Emit |
/// |     +-------+    +----+----+    +------+
/// |                       |
/// |                +------+------+
/// |                |             |
/// |           +----v----+   +----v----+
/// |           | Event A |   | Event B |
/// |           +----+----+   +----+----+
/// |                |             |
/// |         +------+------+      +-------+
/// |         |             |              |
/// |  +------v-----+ +-----v------+ +-----+------+
/// |  | Listener 1 | | Listener 2 | | Listener 3 |
/// |  +------+-----+ +-----+------+ +------------+
/// |         |             |
/// +---------+-------------+
/// ```
library omnibus;

import 'dart:async';
import 'dart:collection';

/// # Omnibus data
///
/// All data objects that is transferred trough the Omnibus have to use this
/// class as base class. You are have to extend this class and specify a [type].
/// All public data should be stored in the [data] field. This way other
/// languages than Dart will be able to connect to the Omnibus using a simple
/// mediator.
/// To distinguish between all the different data objects in the Omnibus, all
/// objects that use the same data structure and convey the same kind of data
/// have the same [type].
class OmnibusData {
  /// # Type identifier
  ///
  /// This is a unique identifier for the type of object that is contained in
  /// [data]. Type identifiers can follow any convention, as long as they are
  /// unique. It is recommended to define all type identifiers that are used
  /// in a specific package in one place to prevent collisions. Additionally,
  /// you can use the package name as prefix to prevent collisions with other
  /// packages. A recommended convention: `package.OriginEntity.originMethod`
  static final String type = 'default';

  String get instanceType => type;
}

/// OmnibusEvent is an alias for OmnibusData and should be used as base class by
/// event classes.
class OmnibusEvent extends OmnibusData {}

/// OmnibusReply is an alias for OmnibusData and should be used as base class by
/// reply classes.
class OmnibusReply extends OmnibusData {}

/// # Omnibus listener
///
/// Each instance of this class defines a separate listener. A listener is
/// connected to a channel in the Omnibus. You cannot connect the same listener
/// to multiple channels.
/// All incoming events are first probed using the [probe] function. If the
/// event is of any interest, [probe] returns true. The event is then passed on
/// to the [receive] function. The [receive] function can optionally send a
/// reply to the event sender by returning a non-null value.
abstract class OmnibusListener {
  /// Tester to only receive a subset of all emitted events. This function
  /// should be very small.
  bool probe(OmnibusEvent event);

  /// Event handler which is called if [probe] returns true. This function
  /// should be programmed asynchronous.
  Future<OmnibusReply> receive(OmnibusEvent event);
}

/// # Omnibus controller
///
/// The Omnibus controller contains a HashMap of listener sets. Each set
/// represents the listeners for a different [OmnibusData.type].
class Omnibus {
  /// Key-value map for all data types.
  HashMap<String, Set<OmnibusListener>> _listeners;

  /// Omnibus constructor
  Omnibus() {
    _listeners = new HashMap<String, Set<OmnibusListener>>();
  }

  /// Add OminbusListener.
  void subscribe(String channel, OmnibusListener listener) {
    // Add new Set of listeners if absent for this OmnibusEvent.
    _listeners.putIfAbsent(channel, () => new Set<OmnibusListener>());

    // Add this listener.
    _listeners[channel].add(listener);
  }

  /// Emit OmnibusEvent.
  Future<Set<OmnibusReply>> emit(final OmnibusEvent event,
      [final List<String> expect]) async {
    var completer = new Completer<Set<OmnibusReply>>();

    // Create Set for return value.
    var replies = new Set<OmnibusReply>();

    // Probe all listeners for this event.
    int listenerCount = _listeners[event.instanceType].length;
    _listeners[event.instanceType].forEach((OmnibusListener listener) {
      // Proceed if the listener wants to receive this event.
      if (listener.probe(event)) {
        // Get reply from listener.
        listener.receive(event).then((OmnibusReply reply) {
          replies.add(reply);
          listenerCount--;
          if (listenerCount == 0) {
            completer.complete(replies);
          }
        });
      }
    });

    // Return collected replies.
    return completer.future;
  }
}
